import re

import pytz
from dateutil.parser import parser
from dli.table import Table

from dli.column import Column
from dli.table_storage_type import TableStorageTypes
from dli.table_types import TableTypes


class HiveSerDe:
    def __init__(self, serde, input_format, output_format):
        self.serde = serde
        self.input_format = input_format
        self.output_format = output_format

    def get_serde(self):
        return self.serde

    def get_input_format(self):
        return self.input_format

    def get_output_format(self):
        return self.output_format


HIVE_SER_DE_MAP = {
    "parquet": HiveSerDe(
        "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
        "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
        "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat"
    ),
    "avro": HiveSerDe(
        "org.apache.hadoop.hive.serde2.avro.AvroSerDe",
        "org.apache.hadoop.hive.ql.io.avro.AvroContainerInputFormat",
        "org.apache.hadoop.hive.ql.io.avro.AvroContainerOutputFormat"
    ),
    "orc": HiveSerDe(
        "org.apache.hadoop.hive.ql.io.orc.OrcSerde",
        "org.apache.hadoop.hive.ql.io.orc.OrcInputFormat",
        "org.apache.hadoop.hive.ql.io.orc.OrcOutputFormat"
    ),
    "rcfile": HiveSerDe(
        "org.apache.hadoop.hive.serde2.columnar.LazyBinaryColumnarSerDe",
        "org.apache.hadoop.hive.ql.io.RCFileOutputFormat",
        "org.apache.hadoop.hive.ql.io.RCFileOutputFormat"
    ),
    "sequencefile": HiveSerDe(
        "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
        "org.apache.hadoop.mapred.SequenceFileInputFormat",
        "org.apache.hadoop.mapred.SequenceFileOutputFormat"
    ),
    "textfile": HiveSerDe(
        "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe",
        "org.apache.hadoop.mapred.TextInputFormat",
        "org.apache.hadoop.hive.ql.io.HiveIgnoreKeyTextOutputFormat"
    )
}


def parse_hetu_show_table_result(database, table, rows):
    table = Table(table, database)
    columns = []
    for row in rows:
        first_col_v = row[0]
        if first_col_v is None or len(first_col_v) == 0:
            continue
        second_v = row[2]
        if second_v is not None and second_v == "partition key":
            columns.append(Column(first_col_v, row[1], row[3], True))
        else:
            columns.append(Column(first_col_v, row[1], row[3]))
    table.columns = columns
    return table


def parse_spark_show_table_result(database, table, rows):
    common_column = True
    partition_column = None
    table = Table(table, database)
    columns = []

    provider = None
    serde_lib = None
    input_format = None
    output_format = None
    not_partition_columns = {}
    for row in rows:
        first_col_v = row[0]
        if common_column:
            if first_col_v == "# Partition Information":
                common_column = False
                partition_column = True
                continue
            elif first_col_v is None or len(first_col_v) == 0:
                common_column = False
                continue
            item = Column(first_col_v, row[1], row[2])
            columns.append(item)
            not_partition_columns[first_col_v] = item
        elif partition_column:
            if first_col_v == "# col_name":
                continue
            elif first_col_v is None or len(first_col_v) == 0:
                partition_column = False
                continue

            item = not_partition_columns.get(first_col_v, None)
            if item is not None:
                item._is_partition_column = True
            else:
                columns.append(Column(first_col_v, row[1], row[2], True))
        else:
            _init_table_desc_info(table, row)
            first_col = row[0]
            second_col = row[1]
            if first_col == "Provider":
                provider = second_col
            elif first_col == "Serde Library":
                serde_lib = second_col
            elif first_col == "InputFormat":
                input_format = second_col
            elif first_col == "OutputFormat":
                output_format = second_col

    if provider == "hive":
        table.data_type = _get_file_format(serde_lib, input_format, output_format)
    table.columns = columns
    return table


def _init_table_desc_info(table, row):
    first_col = row[0]
    second_col = row[1]
    if first_col == "Owner":
        table.owner = second_col
    elif first_col == "Created Time" and second_col is not None and second_col.lower != "unknown":
        table.create_time = _time_string_to_long(second_col)
    elif first_col == "Last Access" and second_col is not None and second_col.lower != "unknown":
        table.last_access_time = _time_string_to_long(second_col)
    elif first_col == "Provider":
        if second_col != "hive":
            table.data_type = second_col
    elif first_col == "Location":
        table.location = second_col
    elif first_col == "Type":
        table.table_type = _get_table_type(second_col)
    elif first_col == "Comment":
        table.description = second_col
    elif first_col == "Table Properties":
        table.table_size = int(re.findall(r"dataSize=(\d+)", second_col)[0]) if re.findall(r"dataSize=(\d+)",
                                                                                          second_col) else 0
    else:
        pass

def _get_table_type(table_type):
    name = table_type.upper()
    if name == TableTypes.EXTERNAL.value or name == "EXTERNAL_TABLE":
        return TableTypes.EXTERNAL.value
    elif name == TableTypes.MANAGED.value or name == "MANAGED_TABLE":
        return TableTypes.MANAGED.value
    elif name == TableTypes.VIEW.value or name == "VIRTUAL_VIEW":
        return TableTypes.VIEW.value
    else:
        return "Unknown"


def _time_string_to_long(time_str):
    if time_str is None or len(time_str) == 0:
        return 0
    try:
        dt = parser().parse(timestr=time_str)
        dt_utc = dt.astimezone(pytz.UTC)
        return int(dt_utc.timestamp() * 1000)
    except Exception as e:
        return None



def _get_file_format(serde, input_format, output_format):
    for key, hive_ser_de in HIVE_SER_DE_MAP.items():
        if (hive_ser_de.get_serde() == serde and
                hive_ser_de.get_input_format() == input_format and
                hive_ser_de.get_output_format() == output_format):
            return key
    return "Unknown"


def create_table_sql(
        database,
        table,
        columns,
        comment=None,
        storage_type=TableStorageTypes.PARQUET.value,
        location=None,
        tbl_properties=None):
    cols = []
    for column in columns:
        if column.is_partition_column:
            continue
        col_str = column.name + " " + column.type
        if column.description is not None and len(column.description) > 0:
            col_str += " COMMENT " + "'" + column.description + "'"
        cols.append(col_str)

    pt_cols = []
    for column in columns:
        if not column.is_partition_column:
            continue
        col_str = column.name + " " + column.type
        if column.description is not None and len(column.description) > 0:
            col_str += " COMMENT " + "'" + column.description + "'"
        pt_cols.append(col_str)

    sql = "CREATE TABLE IF NOT EXISTS `{}`.`{}`({}) STORED AS {}".format(database, table, ",".join(cols), storage_type)
    if comment is not None and len(comment) > 0:
        sql += " COMMENT '{}'".format(comment)
    if location is not None:
        sql += " LOCATION '{}'".format(location)
    if pt_cols is not None and len(pt_cols) > 0:
        sql += " PARTITIONED BY ({})".format(",".join(pt_cols))
    if tbl_properties is not None:
        props = []
        for key, value in tbl_properties.items():
            props.append("{}='{}'".format(key, value))
        sql += " TBLPROPERTIES({})".format(",".join(props))
    return sql
