#!/usr/bin/python
# -*- coding: UTF-8 -*-

from __future__ import absolute_import
from __future__ import unicode_literals

from collections import deque
from decimal import Decimal
from dli.dli_client import DliClient
from dli.dli_sql_client import DliSqlClient
from dli.exception import DliException
from dli.logger import logger
from dli.sql_engine import SqlEngine
from pyDLI import common
from pyDLI.exc import ProgrammingError, OperationalError
from six.moves.urllib.parse import urlparse

import datetime
import re

_TIMESTAMP_PATTERN = re.compile(r'(\d+-\d+-\d+ \d+:\d+:\d+(\.\d{,6})?)')
apilevel = "2.0"
threadsafety = 2  # Threads may share the module and connections.
paramstyle = "named"


class Error(Exception):
    pass


def connect(*args, **kwargs):
    """Constructor for creating a connection to the database. See class :py:class:`Connection` for
    arguments.
    :returns: a :py:class:`Connection` object.
    """
    return Connection(*args, **kwargs)


def _parse_timestamp(value):
    if value:
        match = _TIMESTAMP_PATTERN.match(value)
        if match:
            if match.group(2):
                ts_format = '%Y-%m-%d %H:%M:%S.%f'
                # use the pattern to truncate the value
                value = match.group()
            else:
                ts_format = '%Y-%m-%d %H:%M:%S'
            value = datetime.datetime.strptime(value, ts_format)
        else:
            raise Exception(
                'Cannot convert "{}" into a datetime'.format(value))
    else:
        value = None
    return value


TYPES_CONVERTER = {"DECIMAL_TYPE": Decimal,
                   "TIMESTAMP_TYPE": _parse_timestamp}


class DLIParamEscaper(common.ParamEscaper):
    def escape_string(self, item):
        # backslashes and single quotes need to be escaped
        # TODO verify against parser
        # Need to decode UTF-8 because of old sqlalchemy.
        # Newer SQLAlchemy checks dialect.supports_unicode_binds before encoding Unicode strings
        # as byte strings. The old version always encodes Unicode as byte strings, which breaks
        # string formatting here.
        if isinstance(item, bytes):
            item = item.decode('utf-8')
        return "'{}'".format(
            item
            .replace('\\', '\\\\')
            .replace("'", "\\'")
            .replace('\r', '\\r')
            .replace('\n', '\\n')
            .replace('\t', '\\t')
        )


_escaper = DLIParamEscaper()


def connect(*args, **kwargs):
    """Constructor for creating a connection to the database. See class :py:class:`Connection` for
    arguments.

    :returns: a :py:class:`Connection` object.
    """
    return Connection(*args, **kwargs)


class Connection(object):
    """Wraps a Thrift session"""

    def __init__(self, region=None, endpoint=None, project_id=None, ak=None, sk=None, security_token=None, database=None, queue=None,
                 catalog=None, engine=SqlEngine.SPARK.value, retry_times=1000, obs_endpoint=None):
        self.queue_name = "default"
        if queue is not None:
            self.queue_name = queue
        self.database = "default"
        if database is not None:
            self.database = database
        self.catalog = catalog
        self.engine = engine
        self.retry_times = retry_times
        logger.info("connecting DLI on %s, project %s", region, project_id)
        kwargs = {
            'ak': ak,
            'sk': sk,
            'security_token':security_token,
            'region': region,
            'endpoint': endpoint,
            'project_id': project_id,
            'obs_endpoint': obs_endpoint
        }
        self.dli_client = DliClient(**kwargs)
        self.dli_sql_client = DliSqlClient(self.dli_client)

    def cursor(self, *args, **kwargs):
        return Cursor(self.dli_client, self.queue_name, self.database, catalog=self.catalog,
                      engine=self.engine, retry_times=self.retry_times,
                      *args, **kwargs)

    def close(self):
        # there is no long polling for dli
        # do nothing for close
        pass

    def commit(self):
        pass


class Cursor(common.DBAPICursor):
    """These objects represent a database cursor, which is used to manage the context of a fetch
        operation.

        Cursors are not isolated, i.e., any changes done to the database by a cursor are immediately
        visible by other cursors or connections.
        """

    def __init__(self, dli_client, queue_name, database, catalog, engine, retry_times=1000, arraysize=1000):
        super(Cursor, self).__init__()
        self._dli_client = dli_client
        self._queue_name = queue_name
        self._database = database
        self._catalog = catalog
        self._engine = engine
        self._retry_times = retry_times
        self._arraysize = arraysize
        self._sql_job = None
        self._description = None
        self._state = None
        self._result_set = None
        self._data = deque([])

    def _reset_state(self):
        pass

    @property
    def arraysize(self):
        return self._arraysize

    @arraysize.setter
    def arraysize(self, value):
        """Array size cannot be None, and should be an integer"""
        default_arraysize = 1000
        try:
            self._arraysize = int(value) or default_arraysize
        except TypeError:
            self._arraysize = default_arraysize

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result column:

        - name
        - type_code
        - display_size (None in current implementation)
        - internal_size (None in current implementation)
        - precision (None in current implementation)
        - scale (None in current implementation)
        - null_ok (always True in current implementation)

        This attribute will be ``None`` for operations that do not return rows or if the cursor has
        not had an operation invoked via the :py:meth:`execute` method yet.

        The ``type_code`` can be interpreted by comparing it to the Type Objects specified in the
        section below.
        """
        if self._state == self._STATE_NONE:
            raise ProgrammingError("No query yet")
        if self._description is None:
            columns = self._sql_job.get_schema(queue_name=self._queue_name)
            self._description = []
            for column in columns:
                self._description.append((column.name, column.type, None, None, None, None, True))
        return self._description

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._reset_state()

    def execute(self, operation, parameters=None, options=None, **kwargs):
        """
        Prepare and execute a database operation (query or command).

        :param str operation: SQL statement
        :param sequence or mapping parameters: Input parameters which will be bound to variables in the operation.
        :param dict options: User-defined conf that applies to the sql job
        :return: SqlJob
        """
        # backward compatibility with Python < 3.7
        for kw in ['async', 'async_']:
            if kw in kwargs:
                async_ = kwargs[kw]
                break
        else:
            async_ = False

        # Prepare statement
        if parameters is None:
            sql = operation
        else:
            sql = operation % _escaper.escape_args(parameters)
        logger.debug("execute sql: %s", sql)

        self._reset_state()
        self._state = self._STATE_RUNNING
        if async_:
            self._sql_job = self._dli_client.async_execute_sql(sql=sql,
                                                               db_name=self._database,
                                                               queue_name=self._queue_name,
                                                               catalog=self._catalog,
                                                               engine=self._engine,
                                                               options=options)
            logger.info("async submit job: %s", self._sql_job.job_id)
        else:
            self._sql_job = self._dli_client.execute_sql(sql=sql,
                                                         db_name=self._database,
                                                         queue_name=self._queue_name,
                                                         catalog=self._catalog,
                                                         engine=self._engine,
                                                         options=options)
            logger.info("submit job: %s", self._sql_job.job_id)
            _check_status(self._sql_job)

    def cancel(self):
        if self._state == self._STATE_NONE:
            raise ProgrammingError("No query yet")
        if self._sql_job.job_status in ('FINISHED', 'FAILED'):
            raise ProgrammingError("The job has finished, can not be canceled")
        if self._sql_job.job_status == 'CANCELLED':
            raise ProgrammingError("The job has canceled")
        try:
            self._sql_job.cancel_job()
        except DliException as e:
            raise ProgrammingError("Cancel job failed, reason: " + e.msg)
        return True

    def _fetch_more(self):
        assert (self._state == self._STATE_RUNNING), "Should be running when in _fetch_more"
        logger.info("starting fetch all job result from DLI")
        result_set = self._sql_job.get_result(page_size=self._arraysize, queue_name=self._queue_name)
        while result_set.next():
            new_data = result_set.fetch().columns
            self._data.append(list(new_data))
        logger.info("finished fetch job result")
        self._state = self._STATE_FINISHED

    def poll(self, unused=None):
        """Poll for and return the raw status data provided by the DLI REST API.
        :returns: status in string
        :raises: ``ProgrammingError`` when no query has been started
        .. note::
            This is not a part of DB-API.
        """
        if self._state == self._STATE_NONE:
            raise ProgrammingError("No query yet")
        return self._sql_job.get_status()

    def fetch_logs(self):
        if self._state == self._STATE_NONE:
            raise ProgrammingError("No query yet")
        logger.warn("we do not support to fetch logs")
        return []


def _check_status(sql_job):
    if sql_job.job_status not in ('FINISHED', 'FAILED', 'CANCELLED'):
        logger.error("%s is not a terminate status")
        raise OperationalError(sql_job)
