import contextlib
import threading
import time

import six
from dli.dli_client import DliClient
from sqlalchemy import types as sa_types
from sqlalchemy.engine import Engine, default
from sqlalchemy.sql import compiler, sqltypes

from dli.dli_sql_client import DliSqlClient
from dli.exception import DliException
from dli.sql_engine import SqlEngine
from dli.types import DLITypes

try:
    from sqlalchemy.dialects import mysql
except ImportError:
    # for low sqlalchemy versions
    from sqlalchemy.databases import mysql

_dli_type_to_sqlalchemy_type = {
    DLITypes.BOOLEAN.value: sa_types.Boolean,
    DLITypes.BYTE.value: mysql.TINYINT,
    DLITypes.TINYINT.value: mysql.TINYINT,
    DLITypes.SMALLINT.value: sa_types.SmallInteger,
    DLITypes.SHORT.value: sa_types.SmallInteger,
    DLITypes.INT.value: sa_types.Integer,
    DLITypes.INTEGER.value: sa_types.Integer,
    DLITypes.BIGINT.value: sa_types.BigInteger,
    DLITypes.LONG.value: sa_types.BigInteger,
    DLITypes.FLOAT.value: sa_types.Float,
    DLITypes.REAL.value: sa_types.Float,
    DLITypes.DOUBLE.value: sa_types.Float,
    DLITypes.STRING.value: sa_types.String,
    DLITypes.VARCHAR.value: sa_types.String,
    DLITypes.CHAR.value: sa_types.String,
    DLITypes.DATE.value: sa_types.Date,
    DLITypes.TIME.value: sa_types.TIMESTAMP,
    DLITypes.TIMESTAMP.value: sa_types.TIMESTAMP,
    DLITypes.BINARY.value: sa_types.String,
    DLITypes.VARBINARY.value: sa_types.String,
    DLITypes.ARRAY.value: sa_types.String,
    DLITypes.MAP.value: sa_types.String,
    DLITypes.STRUCT.value: sa_types.String,
    DLITypes.ROW.value: sa_types.String,
    DLITypes.DECIMAL.value: sa_types.DECIMAL,
}

_sqlalchemy_global_reusable_dli = {}

_sqlalchemy_obj_list_cache = {}


class ObjectCache(object):
    def __init__(self, expire=24 * 3600):
        self._expire_time = expire
        self._items = dict()
        self._cache_time = dict()

    def __getitem__(self, key):
        if self._cache_time[key] < time.time() - self._expire_time:
            self._cache_time.pop(key, None)
            self._items.pop(key, None)
            raise KeyError(key)
        return self._items[key]

    def __setitem__(self, key, value):
        self._items[key] = value
        self._cache_time[key] = time.time()

    def get(self, key, default_value=None):
        try:
            return self[key]
        except KeyError:
            return default_value


class DLIIdentifierPreparer(compiler.IdentifierPreparer):
    # Just quote everything to make things simpler / easier to upgrade
    reserved_words = compiler.RESERVED_WORDS.copy()
    keywords = [
        "ADD",
        "ALL",
        "ALTER",
        "AND",
        "AS",
        "ASC",
        "BETWEEN",
        "BIGINT",
        "BOOLEAN",
        "BY",
        "CASE",
        "CAST",
        "COLUMN",
        "COMMENT",
        "CREATE",
        "DESC",
        "DISTINCT",
        "DISTRIBUTE",
        "DOUBLE",
        "DROP",
        "ELSE",
        "FALSE",
        "FROM",
        "FULL",
        "GROUP",
        "IF",
        "IN",
        "INSERT",
        "INTO",
        "IS",
        "JOIN",
        "LEFT",
        "LIFECYCLE",
        "LIKE",
        "LIMIT",
        "MAPJOIN",
        "NOT",
        "NULL",
        "ON",
        "OR",
        "ORDER",
        "OUTER",
        "OVERWRITE",
        "PARTITION",
        "RENAME",
        "REPLACE",
        "RIGHT",
        "RLIKE",
        "SELECT",
        "SORT",
        "STRING",
        "TABLE",
        "TABLESAMPLE",
        "TBLPROPERTIES",
        "THEN",
        "TOUCH",
        "TRUE",
        "UNION",
        "VIEW",
        "WHEN",
        "WHERE",
    ]
    reserved_words.update(keywords)
    reserved_words.update([s.lower() for s in keywords])

    def __init__(self, dialect):
        super(DLIIdentifierPreparer, self).__init__(
            dialect,
            initial_quote="`",
            escape_quote="`",
        )

    def quote(self, ident, force=None):
        return to_str(super(DLIIdentifierPreparer, self).quote(ident, force=force))


class DLICompiler(compiler.SQLCompiler):
    def visit_column(self, *args, **kwargs):
        result = super(DLICompiler, self).visit_column(*args, **kwargs)
        dot_count = result.count(".")


        if dot_count not in (0, 1, 2):
            raise DliException(msg="Unexpected visit_column result {}".format(result))
        if dot_count == 2:
            # we have something of the form schema.table.column
            # hive doesn't like the schema in front, so chop it out
            result = result[result.index(".") + 1:]
        return result

    def visit_char_length_func(self, fn, **kw):
        return "length{}".format(self.function_argspec(fn, **kw))

    def __unicode__(self):
        return to_text(self)


class DLITypeCompiler(compiler.GenericTypeCompiler):
    def visit_INTEGER(self, type_):
        return "INT"

    def visit_NUMERIC(self, type_):
        return "DECIMAL"

    def visit_CHAR(self, type_):
        return "STRING"

    def visit_VARCHAR(self, type_):
        return "STRING"

    def visit_NCHAR(self, type_):
        return "STRING"

    def visit_TEXT(self, type_):
        return "STRING"

    def visit_CLOB(self, type_):
        return "STRING"

    def visit_BLOB(self, type_):
        return "BINARY"

    def visit_TIME(self, type_):
        return "TIMESTAMP"

    def visit_DATETIME(self, type_, **kw):
        return "STRING"


if hasattr(sqltypes.String, "RETURNS_UNICODE"):
    _return_unicode_str = sqltypes.String.RETURNS_UNICODE
else:
    _return_unicode_str = True


class DLIDialect(default.DefaultDialect):
    name = "dli"
    driver = "dli"
    preparer = DLIIdentifierPreparer
    statement_compiler = DLICompiler
    supports_views = True
    supports_alter = True
    supports_pk_autoincrement = False
    supports_default_values = False
    supports_empty_insert = False
    supports_native_decimal = True
    supports_native_boolean = True
    supports_unicode_statements = True
    supports_unicode_binds = True
    returns_unicode_strings = _return_unicode_str
    description_encoding = None
    supports_multivalues_insert = True
    type_compiler = DLITypeCompiler
    supports_sane_rowcount = False
    supports_statement_cache = False
    default_schema_name = "default"

    @classmethod
    def dbapi(cls):
        from . import dli

        return dli

    def create_connect_args(self, url):
        url_string = str(url)
        region = url.host
        access_id = url.username
        secret_access_key = url.password
        project = None
        dli_endpoint = None
        obs_endpoint = None
        queue = None
        database = None
        catalog = None
        engine = None
        retry_times = None

        cache_names = False
        cache_seconds = 24 * 3600

        if url.query:
            query = dict(url.query)
            project = query.pop("projectid", None)
            dli_endpoint = query.pop("endpoint", None)
            queue = query.pop("queuename", None)
            database = query.pop("databasename", None)
            obs_endpoint = query.pop("obsendpoint", None)
            catalog = query.pop("catalog", None)
            engine = query.pop("enginetype", SqlEngine.SPARK.value)
            retry_times = query.pop("retrytimes", 100)

            if cache_names is False:
                cache_names = query.pop("cache_names", "false").lower() != "false"
            cache_seconds = int(query.pop("cache_seconds", cache_seconds))

        kwargs = {
            "region": region,
            "endpoint": dli_endpoint,
            "project_id": project,
            "ak": access_id,
            "sk": secret_access_key,
            "database": database,
            "queue": queue,
            "obs_endpoint": obs_endpoint,
            "catalog": catalog,
            "engine": engine,
            "retry_times": retry_times,
        }

        not_none_kwargs = {
            "region": region,
            "project_id": project,
            "ak": access_id,
            "sk": secret_access_key,
            "database": database,
            "queue": queue,
        }

        for k, v in six.iteritems(not_none_kwargs):
            if v is None:
                raise ValueError(
                    "{} should be provided to create connection, "
                    "you can either specify in connection string as format: "
                    '"dli://<access_id>:<access_key>@<region>", '
                    "or create an dli object and call `.to_global()` "
                    "to set it to global".format(k)
                )

        if cache_names:
            _sqlalchemy_obj_list_cache[url_string] = ObjectCache(expire=cache_seconds)

        return [], kwargs

    @classmethod
    def get_list_cache(cls, url, key):
        url = str(url)
        if url not in _sqlalchemy_obj_list_cache:
            return None
        return _sqlalchemy_obj_list_cache[url].get(key)

    @classmethod
    def put_list_cache(cls, url, key, value):
        url = str(url)
        if url not in _sqlalchemy_obj_list_cache:
            return
        _sqlalchemy_obj_list_cache[url][key] = value

    def _get_dli_client(self, **kwargs):
        param = {
            'ak': kwargs.pop("ak"),
            'sk': kwargs.pop("sk"),
            'region': kwargs.pop("region"),
            'endpoint': kwargs.pop("endpoint"),
            'project_id': kwargs.pop("project_id"),
            'obs_endpoint': kwargs.pop("obs_endpoint")
        }
        return DliSqlClient(DliClient(**param))

    def get_schema_names(self, connection, **kw):
        conn = self._get_dbapi_connection(connection)
        try:
            return [schema for schema in conn.dli_sql_client.list_databases(queue=conn.queue_name,
                                                                            catalog=conn.catalog,
                                                                            engine=conn.engine)]
        except DliException:
            return ["default"]

    def has_table(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        return conn.dli_sql_client.table_exists(database=schema,
                                                table=table_name,
                                                queue=conn.queue_name,
                                                catalog=conn.catalog,
                                                engine=conn.engine)

    @classmethod
    def _get_dbapi_connection(cls, sa_connection):
        if isinstance(sa_connection, Engine):
            sa_connection = sa_connection.connect()
        return sa_connection.connection.connection

    @classmethod
    def _get_schema_kw(cls, connection, schema=None):
        return schema

    def get_columns(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        table = conn.dli_sql_client.get_table(database=schema,
                                              table=table_name,
                                              queue=conn.queue_name,
                                              catalog=conn.catalog,
                                              engine=conn.engine)
        result = []
        for col in table.columns:
            col_type = _dli_type_to_sqlalchemy_type.get(col.type, None)
            if col_type is None:
                if col.type.startswith(DLITypes.DECIMAL.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.DECIMAL.value)
                elif col.type.startswith(DLITypes.CHAR.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.CHAR.value)
                elif col.type.startswith(DLITypes.VARCHAR.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.VARCHAR.value)
                elif col.type.startswith(DLITypes.ARRAY.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.ARRAY.value)
                elif col.type.startswith(DLITypes.MAP.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.MAP.value)
                elif col.type.startswith(DLITypes.STRUCT.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.STRUCT.value)
                elif col.type.startswith(DLITypes.ROW.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.ROW.value)
                elif col.type.startswith(DLITypes.TIMESTAMP.value):
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.TIMESTAMP.value)
                else:
                    col_type = _dli_type_to_sqlalchemy_type.get(DLITypes.STRING.value)
            result.append(
                {
                    "name": col.name,
                    "type": col_type,
                    "nullable": True,
                    "default": None,
                    "comment": col.description,
                }
            )
        return result

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        return []

    def get_indexes(self, connection, table_name, schema=None, **kw):
        return []

    def get_table_names(self, connection, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        return conn.dli_sql_client.list_tables(database=schema,
                                               queue=conn.queue_name,
                                               catalog=conn.catalog,
                                               engine=conn.engine)

    def get_view_names(self, connection, schema=None, **kw):
        return []

    def get_table_comment(self, connection, table_name, schema=None, **kw):
        conn = self._get_dbapi_connection(connection)
        comment = conn.dli_sql_client.get_table(database=schema,
                                                table=table_name,
                                                queue=conn.queue_name,
                                                catalog=conn.catalog,
                                                engine=conn.engine).description
        return {"text": comment}

    @classmethod
    def _is_stack_superset(cls, tb):
        try:
            cur_frame = tb.tb_frame
            while cur_frame is not None:
                if "superset" in cur_frame.f_code.co_filename:
                    return True
                cur_frame = cur_frame.f_back
            return False
        except DliException:  # pragma: no cover
            return False

    def do_rollback(self, dbapi_connection):
        pass

    def _check_unicode_returns(self, connection, additional_tests=None):
        # We decode everything as UTF-8
        return True

    def _check_unicode_description(self, connection):
        # We decode everything as UTF-8
        return True


def to_text(binary, encoding="utf-8"):
    if binary is None:
        return binary
    if isinstance(binary, (six.binary_type, bytearray)):
        return binary.decode(encoding)
    elif isinstance(binary, six.text_type):
        return binary
    else:
        return str(binary) if six.PY3 else str(binary).decode(encoding)


def to_str(text, encoding="utf-8"):
    return (
        to_text(text, encoding=encoding)
        if six.PY3
        else to_binary(text, encoding=encoding)
    )


def to_binary(text, encoding="utf-8"):
    if text is None:
        return text
    if isinstance(text, six.text_type):
        return text.encode(encoding)
    elif isinstance(text, (six.binary_type, bytearray)):
        return bytes(text)
    else:
        return str(text).encode(encoding) if six.PY3 else str(text)
